from Dictionary.TxtWord import TxtWord
from Dictionary.Word import Word
from Language.TurkishLanguage import TurkishLanguage


class MorphotacticEngine:

    @staticmethod
    def resolveD(root: TxtWord,
                 formation: str,
                 formationToCheck: str) -> str:
        """
        resolveD resolves the D metamorpheme to 'd' or 't' depending on the root and current formationToCheck. It adds
        'd' if the root is an abbreviation; 't' if the last phoneme is one of the "çfhkpsşt" (fıstıkçı şahap) or 'd'
        otherwise; 't' if the word is a number ending with 3, 4, 5, 40, 60, or 70 or 'd' otherwise.
        :param root: Root of the word
        :param formation: Formation is current status of the wordform in the current state of the finite state machine. It
                          is always equal to formationToCheck except the case where there is an apostrophe after the
                          formationToCheck such as (3').
        :param formationToCheck: FormationToCheck is current status of the wordform in the current state of the finite
                                 state machine except the apostrophe at the end if it exists.
        :return: Formation with added 'd' or 't' character.
        """
        if root.isAbbreviation():
            return formation + 'd'
        if "0" <= Word.lastPhoneme(formationToCheck) <= "9":
            if Word.lastPhoneme(formationToCheck) == "3" or Word.lastPhoneme(formationToCheck) == "4"\
                    or Word.lastPhoneme(formationToCheck) == "5":
                return formation + 't'
            elif Word.lastPhoneme(formationToCheck) == "0":
                if root.getName().endswith("40") or root.getName().endswith("60") or root.getName().endswith("70"):
                    return formation + 't'
                else:
                    return formation + 'd'
            else:
                return formation + 'd'
        else:
            if TurkishLanguage.isSertSessiz(Word.lastPhoneme(formationToCheck)):
                return formation + 't'
            else:
                return formation + 'd'

    @staticmethod
    def resolveA(root: TxtWord,
                 formation: str,
                 rootWord: bool,
                 formationToCheck: str):
        """
        resolveA resolves the A metamorpheme to 'a' or 'e' depending on the root and current formationToCheck. It adds
        'e' if the root is an abbreviation; 'a' if the last vowel is a back vowel (except words that do not obey vowel
        harmony during agglutination); 'e' if the last vowel is a front vowel (except words that do not obey vowel
        harmony during agglutination); 'a' if the word is a number ending with 6, 9, 10, 30, 40, 60, or 90 or 'e'
        otherwise.
        :param root: Root of the word
        :param formation: Formation is current status of the wordform in the current state of the finite state machine. It
                          is always equal to formationToCheck except the case where there is an apostrophe after the
                          formationToCheck such as (3').
        :param rootWord: True if the current word form is root form, false otherwise.
        :param formationToCheck: FormationToCheck is current status of the wordform in the current state of the finite
                                 state machine except the apostrophe at the end if it exists.
        :return: Formation with added 'a' or 'e' character.
        """
        if root.isAbbreviation():
            return formation + 'e'
        if "0" <= Word.lastVowel(formationToCheck) <= "9":
            if Word.lastVowel(formationToCheck) == "6" or Word.lastVowel(formationToCheck) == "9":
                return formation + 'a'
            elif Word.lastVowel(formationToCheck) == "0":
                if root.getName().endswith("10") or root.getName().endswith("30") or root.getName().endswith("40") \
                        or root.getName().endswith("60") or root.getName().endswith("90"):
                    return formation + 'a'
                else:
                    return formation + 'e'
            else:
                return formation + 'e'
        if TurkishLanguage.isBackVowel(Word.lastVowel(formationToCheck)):
            if root.notObeysVowelHarmonyDuringAgglutination() and rootWord:
                return formation + 'e'
            else:
                return formation + 'a'
        if TurkishLanguage.isFrontVowel(Word.lastVowel(formationToCheck)):
            if root.notObeysVowelHarmonyDuringAgglutination() and rootWord:
                return formation + 'a'
            else:
                return formation + 'e'
        if root.isNumeral() or root.isFraction() or root.isReal():
            if root.getName().endswith("6") or root.getName().endswith("9") or root.getName().endswith("10") or \
                    root.getName().endswith("30") or root.getName().endswith("40") or root.getName().endswith("60") \
                    or root.getName().endswith("90"):
                return formation + 'a'
            else:
                return formation + 'e'
        return formation

    @staticmethod
    def resolveH(root: TxtWord,
                 formation: str,
                 beginningOfSuffix: bool,
                 specialCaseTenseSuffix: bool,
                 rootWord: bool,
                 formationToCheck: str):
        """
        resolveH resolves the H metamorpheme to 'ı', 'i', 'u' or 'ü', depending on the  current formationToCheck, root,
        and formation. It adds 'i' if the root is an abbreviation; 'ü' if the  character before the last vowel is
        front rounded (or back rounded when the root word does not obey vowel harmony during agglutination); 'i' if the
        character before the last vowel is front unrounded; 'u' if the character before the  last vowel is back rounded;
        'ı' if the character before the last vowel is back unrounded (or front unrounded when the root word does not obey
        vowel harmony during agglutination); 'ı' if the word is a  number ending with 6, 40, 60 or 90; 'ü' if the word
        is a number ending with 3, 4, or 00; 'u' if the word is a number ending with 9, 10, or 30; 'i' otherwise for
        numbers. Special case for 'Hyor' suffix is handled with resolveHforSpecialCaseTenseSuffix method.
        :param root: Root of the word
        :param formation: Formation is current status of the wordform in the current state of the finite state machine.
                          It is always equal to formationToCheck except the case where there is an apostrophe after the
                          formationToCheck such as (3').
        :param beginningOfSuffix: True if H appears in the beginning of the suffix, false otherwise.
        :param specialCaseTenseSuffix: True if the suffix is 'Hyor', false otherwise.
        :param rootWord: True if the current word form is root form, false otherwise.
        :param formationToCheck: FormationToCheck is current status of the word form in the current state of the finite
                                 state machine except the apostrophe at the end if it exists.
        :return: Formation with possibly last character dropped and 'ı', 'i', 'u' or 'ü' character added.
        """
        if root.isAbbreviation():
            return formation + 'i'
        if beginningOfSuffix and TurkishLanguage.isVowel(Word.lastPhoneme(formationToCheck)) and \
                not specialCaseTenseSuffix:
            return formation
        if specialCaseTenseSuffix:
            if rootWord:
                if root.vowelAChangesToIDuringYSuffixation():
                    if TurkishLanguage.isFrontRoundedVowel(Word.beforeLastVowel(formationToCheck)):
                        return formation[:len(formation) - 1] + 'ü'
                    if TurkishLanguage.isFrontUnroundedVowel(Word.beforeLastVowel(formationToCheck)):
                        return formation[:len(formation) - 1] + 'i'
                    if TurkishLanguage.isBackRoundedVowel(Word.beforeLastVowel(formationToCheck)):
                        return formation[:len(formation) - 1] + 'u'
                    if TurkishLanguage.isBackUnroundedVowel(Word.beforeLastVowel(formationToCheck)):
                        return formation[:len(formation) - 1] + 'ı'
            if TurkishLanguage.isVowel(Word.lastPhoneme(formationToCheck)):
                if TurkishLanguage.isFrontRoundedVowel(Word.beforeLastVowel(formationToCheck)):
                    return formation[:len(formation) - 1] + 'ü'
                if TurkishLanguage.isFrontUnroundedVowel(Word.beforeLastVowel(formationToCheck)):
                    return formation[:len(formation) - 1] + 'i'
                if TurkishLanguage.isBackRoundedVowel(Word.beforeLastVowel(formationToCheck)):
                    return formation[:len(formation) - 1] + 'u'
                if TurkishLanguage.isBackUnroundedVowel(Word.beforeLastVowel(formationToCheck)):
                    return formation[:len(formation) - 1] + 'ı'
        if TurkishLanguage.isFrontRoundedVowel(Word.lastVowel(formationToCheck)) or \
                (TurkishLanguage.isBackRoundedVowel(Word.lastVowel(formationToCheck))
                 and root.notObeysVowelHarmonyDuringAgglutination()):
            return formation + 'ü'
        if (TurkishLanguage.isFrontUnroundedVowel(Word.lastVowel(formationToCheck)) and
                (not root.notObeysVowelHarmonyDuringAgglutination() or not rootWord)) or \
                ((Word.lastVowel(formationToCheck) == 'a' or Word.lastVowel(formationToCheck) == 'â') and root.notObeysVowelHarmonyDuringAgglutination()):
            return formation + 'i'
        if TurkishLanguage.isBackRoundedVowel(Word.lastVowel(formationToCheck)):
            return formation + 'u'
        if TurkishLanguage.isBackUnroundedVowel(Word.lastVowel(formationToCheck)) or \
                (TurkishLanguage.isFrontUnroundedVowel(Word.lastVowel(formationToCheck)) and
                                                 root.notObeysVowelHarmonyDuringAgglutination()):
            return formation + 'ı'
        if root.isNumeral() or root.isFraction() or root.isReal():
            if root.getName().endswith("6") or root.getName().endswith("40") or root.getName().endswith("60") \
                    or root.getName().endswith("90"):
                return formation + 'ı'
            else:
                if root.getName().endswith("3") or root.getName().endswith("4") or root.getName().endswith("00"):
                    return formation + 'ü'
                else:
                    if root.getName().endswith("9") or root.getName().endswith("10") or root.getName().endswith("30"):
                        return formation + 'u'
                    else:
                        return formation + 'i'
        if Word.lastVowel(formationToCheck) == '0':
            return formation + 'i'
        return formation

    @staticmethod
    def resolveC(formation: str,
                 formationToCheck: str) -> str:
        """
        The resolveC method takes a str formation as an input. If the last phoneme is on of the "çfhkpsşt", it
        concatenates given formation with 'ç', if not it concatenates given formation with 'c'.

        PARAMETERS
        ----------
        formation : str
            String input.

        RETURNS
        -------
        str
            Resolved String.
        """
        if TurkishLanguage.isSertSessiz(Word.lastPhoneme(formationToCheck)):
            return formation + 'ç'
        else:
            return formation + 'c'

    @staticmethod
    def resolveS(formation: str) -> str:
        """
        The resolveS method takes a str formation as an input. It then concatenates given formation with 's'.

        PARAMETERS
        ----------
        formation : str
            String input.

        RETURNS
        -------
        str
            Resolved String.
        """
        return formation + 's'

    @staticmethod
    def resolveSh(formation: str) -> str:
        """
        The resolveSh method takes a str formation as an input. If the last character is a vowel, it concatenates
        given formation with ş, if the last character is not a vowel, and not 't' it directly returns given formation,
        but if it is equal to 't', it transforms it to 'd'.

        PARAMETERS
        ----------
        formation : str
            String input.

        RETURNS
        -------
        str
            Resolved String.
        """
        if TurkishLanguage.isVowel(formation[len(formation) - 1]):
            return formation + 'ş'
        else:
            if formation[len(formation) - 1] != 't':
                return formation
            else:
                return formation[len(formation) - 1] + 'd'



from Dictionary.TxtWord import TxtWord
from Dictionary.Word import Word
from Language.TurkishLanguage import TurkishLanguage

from MorphologicalAnalysis.FsmParse import FsmParse
from MorphologicalAnalysis.MorphotacticEngine import MorphotacticEngine
from MorphologicalAnalysis.State import State


class Transition:

    __to_state: State
    __with: str
    __with_name: str
    __to_pos: str

    def __init__(self,
                 _with: str,
                 toState=None,
                 withName=None,
                 toPos=None):
        """
        Constructor of Transition class which takes a State, and three str as input. Then it
        initializes toState, with, withName and toPos variables with given inputs.

        PARAMETERS
        ----------
        _with : str
            String input.
        toState : State
            State input.
        withName : str
            String input.
        toPos : str
            String input.
        """
        self.__with = _with
        self.__to_state = toState
        self.__with_name = withName
        self.__to_pos = toPos

    def toState(self) -> State:
        """
        Getter for the toState variable.

        RETURNS
        -------
        State
            toState variable.
        """
        return self.__to_state

    def toPos(self) -> str:
        """
        Getter for the toPos variable.

        RETURNS
        -------
        str
            toPos variable.
        """
        return self.__to_pos

    def transitionPossibleForString(self,
                                    currentSurfaceForm: str,
                                    realSurfaceForm: str) -> bool:
        """
        The transitionPossibleForString method takes two str as inputs; currentSurfaceForm and realSurfaceForm. If the
        length of the given currentSurfaceForm is greater than the given realSurfaceForm, it directly returns true. If
        not, it takes a substring from given realSurfaceForm with the size of currentSurfaceForm. Then checks for the
        characters of with variable.

        If the character of with that makes transition is C, it returns true if the substring contains c or ç.
        If the character of with that makes transition is D, it returns true if the substring contains d or t.
        If the character of with that makes transition is A, it returns true if the substring contains a or e.
        If the character of with that makes transition is K, it returns true if the substring contains k, g or ğ.
        If the character of with that makes transition is other than the ones above, it returns true if the substring
        contains the same character as with.

        PARAMETERS
        ----------
        currentSurfaceForm : str
            String input.
        realSurfaceForm : str
            String input.

        RETURNS
        -------
        bool
            True when the transition is possible according to Turkish grammar, False otherwise.
        """
        if len(currentSurfaceForm) == 0 or len(currentSurfaceForm) >= len(realSurfaceForm):
            return True
        search_string = realSurfaceForm[len(currentSurfaceForm):]
        for ch in self.__with:
            if ch == 'C':
                return 'c' in search_string or 'ç' in search_string
            elif ch == 'D':
                return 'd' in search_string or 't' in search_string
            elif ch == 'c' or ch == 'e' or ch == 'r' or ch == 'p' or ch == 'l' or ch == 'b' or ch == 'd' or ch == 'g' \
                    or ch == 'o' or ch == 'm' or ch == 'v' or ch == 'i' or ch == 'ü' or ch == 'z':
                return ch in search_string
            elif ch == 'A':
                return 'a' in search_string or 'e' in search_string
            elif ch == 'k':
                return 'k' in search_string or 'g' in search_string or 'ğ' in search_string
        return True

    def transitionPossibleForParse(self, currentFsmParse: FsmParse) -> bool:
        """
        The transitionPossibleForParse method takes a FsmParse currentFsmParse as an input. It then checks some special
        cases;

        PARAMETERS
        ----------
        currentFsmParse : FsmParse
            Parse to be checked

        RETURNS
        -------
        bool
            True if transition is possible False otherwise
        """
        if self.__with == "Ar" and currentFsmParse.getSurfaceForm().endswith("l") and \
                currentFsmParse.getWord().getName() != currentFsmParse.getSurfaceForm():
            return False
        return True

    def transitionPossibleForWord(self,
                                  root: TxtWord,
                                  fromState: State) -> bool:
        """
        The transitionPossible method takes root and current parse as inputs. It then checks some special cases.
        :param root: Current root word
        :param fromState: From which state we arrived to this state.
        :return: true if transition is possible false otherwise
        """
        if root.isAdjective() and ((root.isNominal() and not root.isExceptional()) or root.isPronoun()) \
                and self.__to_state.getName() == "NominalRoot(ADJ)" and self.__with == "0":
            return False
        if root.isAdjective() and root.isNominal() and self.__with == "^DB+VERB+ZERO+PRES+A3PL" \
                and fromState.getName() == "AdjectiveRoot":
            return False
        if root.isAdjective() and root.isNominal() and self.__with == "SH" and fromState.getName() == "AdjectiveRoot":
            return False
        if self.__with == "ki":
            return root.takesRelativeSuffixKi()
        if self.__with == "kü":
            return root.takesRelativeSuffixKu()
        if self.__with == "DHr":
            if self.__to_state.getName() == "Adverb":
                return True
            else:
                return root.takesSuffixDIRAsFactitive()
        if self.__with == "Hr" and (
                self.__to_state.getName() == "AdjectiveRoot(VERB)" or self.__to_state.getName() == "OtherTense" or
                self.__to_state.getName() == "OtherTense2"):
            return root.takesSuffixIRAsAorist()
        return True

    def __withFirstChar(self) -> str:
        """
        The withFirstChar method returns the first character of the with variable.

        RETURNS
        -------
        str
            The first character of the with variable.
        """
        if len(self.__with) == 0:
            return "$"
        if self.__with[0] != "'":
            return self.__with[0]
        elif len(self.__with) == 1:
            return self.__with[0]
        else:
            return self.__with[1]

    def __startWithVowelorConsonantDrops(self) -> bool:
        """
        The startWithVowelorConsonantDrops method checks for some cases. If the first character of with variable is
        "nsy", and with variable does not equal to one of the Strings; "ylA, ysA, ymHs, yDH, yken", it returns true. If

        Or, if the first character of with variable is 'A, H: or any other vowels, it returns true.

        RETURNS
        -------
        bool
            True if it starts with vowel or consonant drops, false otherwise.
        """
        if TurkishLanguage.isConsonantDrop(self.__withFirstChar()) and self.__with != "ylA" and self.__with != "ysA" \
                and self.__with != "ymHs" and self.__with != "yDH" and self.__with != "yken":
            return True
        if self.__withFirstChar() == "A" or self.__withFirstChar() == "H" or \
                TurkishLanguage.isVowel(self.__withFirstChar()):
            return True
        return False

    def softenDuringSuffixation(self, root: TxtWord, startState: State) -> bool:
        """
        The startWithVowelorConsonantDrops method checks for some cases. If the first character of with variable is
        "nsy", and with variable does not equal to one of the Strings; "ylA, ysA, ymHs, yDH, yken", it returns true. If

        Or, if the first character of with variable is 'A, H: or any other vowels, it returns true.

        RETURNS
        -------
        bool
            True if it starts with vowel or consonant drops, false otherwise.
        """
        if not startState.getName().startswith("VerbalRoot") and (root.isNominal() or root.isAdjective()) and root.nounSoftenDuringSuffixation() and \
                (self.__with == "Hm" or self.__with == "nDAn" or self.__with == "ncA" or self.__with == "nDA"
                 or self.__with == "yA" or self.__with == "yHm" or self.__with == "yHz" or self.__with == "yH"
                 or self.__with == "nH" or self.__with == "nA" or self.__with == "nHn" or self.__with == "H"
                 or self.__with == "sH" or self.__with == "Hn" or self.__with == "HnHz" or self.__with == "HmHz"):
            return True
        if startState.getName().startswith("VerbalRoot") and root.isVerb() and root.verbSoftenDuringSuffixation() and \
                (self.__with.startswith("Hyor") or self.__with == "yHs" or self.__with == "yAn" or self.__with == "yA"
                 or self.__with == "yAcAk" or self.__with == "yAsH" or self.__with == "yHncA" or self.__with == "yHp"
                 or self.__with == "yAlH" or self.__with == "yArAk" or self.__with == "yAdur" or self.__with == "yHver"
                 or self.__with == "yAgel" or self.__with == "yAgor" or self.__with == "yAbil" or self.__with == "yAyaz"
                 or self.__with == "yAkal" or self.__with == "yAkoy" or self.__with == "yAmA" or self.__with == "yHcH"
                 or self.__with == "yHs" or self.__with == "HCH" or self.__with == "Hr" or self.__with == "Hs"
                 or self.__with == "Hn" or self.__with == "yHn" or self.__with == "yHnHz" or self.__with == "Ar"
                 or self.__with == "Hl"):
            return True
        return False

    def makeTransitionNoStartState(self, root: TxtWord, stem: str) -> str:
        """
        The makeTransition method takes a TxtWord root and s str stem as inputs. If given root is a verb,
        it makes transition with given root and stem with the verbal root state. If given root is not verb, it makes
        transition with given root and stem and the nominal root state.

        PARAMETERS
        ----------
        root : TxtWord
            TxtWord input.
        stem : str
            String input.

        RETURNS
        -------
        str
            String type output that has the transition.
        """
        if root.isVerb():
            return self.makeTransition(root, stem, State("VerbalRoot", True, False))
        else:
            return self.makeTransition(root, stem, State("NominalRoot", True, False))

    def makeTransition(self,
                       root: TxtWord,
                       stem: str,
                       startState: State) -> str:
        """
        The method is main driving method to accomplish the current transition from one state to another depending on
        the root form of the word, current value of the word form, and the type of the start state. The method
        (a) returns the original word form if the transition is an epsilon transition, (b) adds 'nunla' if the current
        stem is 'bu', 'şu' or 'o', (c) returns 'bana' or 'sana' if the current stem is 'ben' or 'sen' respectively.
        For other cases, the method first modifies current stem and then adds the transition using special metamorpheme
        resolving methods. These cases are: (d) Converts 'y' of the first character of the transition to 'i' if the
        current stem is 'ye' or 'de'. (e) Drops the last two characters and adds last character when the transition is
        ('Hl' or 'Hn') and last 'I' drops during passive suffixation. (f) Adds 'y' character when the word ends with 'su'
        and the transition does not start with 'y'. (g) Adds the last character again when the root duplicates during
        suffixation. (h) Drops the last two characters and adds the last character when last 'i' drops during
        suffixation. (i) Replaces the last character with a soft one when the root soften during suffixation.
        :param root: Root of the current word form
        :param stem: Current word form
        :param startState: The state from which this Fsm morphological analysis search has started.
        :return: The current value of the word form after this transition is completed in the finite state machine.
        """
        root_word = root.getName() == stem or (root.getName() + "'") == stem
        formation = stem
        i = 0
        if self.__with == "0":
            return stem
        if (stem == "bu" or stem == "şu" or stem == "o") and root_word and self.__with == "ylA":
            return stem + "nunla"
        if self.__with == "yA":
            if stem == "ben":
                return "bana"
            if stem == "sen":
                return "sana"
        formation_to_check = stem
        if root_word and self.__withFirstChar() == "y" and root.vowelEChangesToIDuringYSuffixation() \
                and (self.__with[1] != "H" or root.getName() == "ye"):
            formation = stem[:len(stem) - 1] + "i"
            formation_to_check = formation
        else:
            if root_word and (self.__with == "Hl" or self.__with == "Hn") and root.lastIdropsDuringPassiveSuffixation():
                formation = stem[:len(stem) - 2] + stem[len(stem) - 1]
                formation_to_check = stem
            else:
                if root_word and root.showsSuRegularities() and self.__startWithVowelorConsonantDrops():
                    formation = stem + 'y'
                    i = 1
                    formation_to_check = formation
                else:
                    if root_word and root.duplicatesDuringSuffixation() and not startState.getName().startswith(
                                    "VerbalRoot") and TurkishLanguage.isConsonantDrop(self.__with[0]):
                        if self.softenDuringSuffixation(root, startState):
                            if Word.lastPhoneme(stem) == "p":
                                formation = stem[:len(stem) - 1] + "bb"
                            elif Word.lastPhoneme(stem) == "t":
                                formation = stem[:len(stem) - 1] + "dd"
                        else:
                            formation = stem + stem[len(stem) - 1]
                        formation_to_check = formation
                    else:
                        if root_word and root.lastIdropsDuringSuffixation() and \
                                not startState.getName().startswith(
                                    "VerbalRoot") and not startState.getName().startswith("ProperRoot") \
                                and self.__startWithVowelorConsonantDrops():
                            if self.softenDuringSuffixation(root, startState):
                                if Word.lastPhoneme(stem) == "p":
                                    formation = stem[:len(stem) - 2] + 'b'
                                elif Word.lastPhoneme(stem) == "t":
                                    formation = stem[:len(stem) - 2] + 'd'
                                elif Word.lastPhoneme(stem) == "ç":
                                    formation = stem[:len(stem) - 2] + 'c'
                            else:
                                formation = stem[:len(stem) - 2] + stem[len(stem) - 1]
                            formation_to_check = stem
                        else:
                            if Word.lastPhoneme(stem) == "p":
                                if self.__startWithVowelorConsonantDrops() and root_word and \
                                        self.softenDuringSuffixation(root, startState):
                                    formation = stem[:len(stem) - 1] + 'b'
                            elif Word.lastPhoneme(stem) == "t":
                                if self.__startWithVowelorConsonantDrops() and root_word and \
                                        self.softenDuringSuffixation(root, startState):
                                    formation = stem[:len(stem) - 1] + 'd'
                            elif Word.lastPhoneme(stem) == "ç":
                                if self.__startWithVowelorConsonantDrops() and root_word and \
                                        self.softenDuringSuffixation(root, startState):
                                    formation = stem[:len(stem) - 1] + 'c'
                            elif Word.lastPhoneme(stem) == "g":
                                if self.__startWithVowelorConsonantDrops() and root_word and \
                                        self.softenDuringSuffixation(root, startState):
                                    formation = stem[:len(stem) - 1] + 'ğ'
                            elif Word.lastPhoneme(stem) == "k":
                                if self.__startWithVowelorConsonantDrops() and root_word and root.endingKChangesIntoG() \
                                        and (not root.isProperNoun() or startState.__str__() != "ProperRoot"):
                                    formation = stem[:len(stem) - 1] + 'g'
                                else:
                                    if self.__startWithVowelorConsonantDrops() and (not root_word or (
                                            self.softenDuringSuffixation(root, startState) and (
                                            not root.isProperNoun() or startState.__str__() != "ProperRoot"))):
                                        formation = stem[:len(stem) - 1] + 'ğ'
                            formation_to_check = formation
        if TurkishLanguage.isConsonantDrop(self.__withFirstChar()) and not TurkishLanguage.isVowel(stem[len(stem) - 1])\
                and (root.isNumeral() or root.isReal() or root.isFraction() or root.isTime() or root.isDate()
                     or root.isPercent() or root.isRange()) \
                and (root.getName().endswith("1") or root.getName().endswith("3") or root.getName().endswith("4")
                     or root.getName().endswith("5") or root.getName().endswith("8") or root.getName().endswith("9")
                     or root.getName().endswith("10") or root.getName().endswith("30") or root.getName().endswith("40")
                     or root.getName().endswith("60") or root.getName().endswith("70") or root.getName().endswith("80")
                     or root.getName().endswith("90") or root.getName().endswith("00")):
            if self.__with[0] == "'":
                formation = formation + "'"
                i = 2
            else:
                i = 1
        else:
            if (TurkishLanguage.isConsonantDrop(self.__withFirstChar()) and TurkishLanguage.isConsonant(
                    Word.lastPhoneme(stem))) or (root_word and root.consonantSMayInsertedDuringPossesiveSuffixation()):
                if self.__with[0] == "'":
                    formation = formation + "'"
                    if root.isAbbreviation():
                        i = 1
                    else:
                        i = 2
                else:
                    i = 1
        while i < len(self.__with):
            if self.__with[i] == "D":
                formation = MorphotacticEngine.resolveD(root, formation, formation_to_check)
            elif self.__with[i] == "A":
                formation = MorphotacticEngine.resolveA(root, formation, root_word, formation_to_check)
            elif self.__with[i] == "H":
                if self.__with[0] != "'":
                    formation = MorphotacticEngine.resolveH(root, formation, i == 0, self.__with.startswith("Hyor"), root_word, formation_to_check)
                else:
                    formation = MorphotacticEngine.resolveH(root, formation, i == 1, False, root_word, formation_to_check)
                root_word = False
            elif self.__with[i] == "C":
                formation = MorphotacticEngine.resolveC(formation, formation_to_check)
            elif self.__with[i] == "S":
                formation = MorphotacticEngine.resolveS(formation)
            elif self.__with[i] == "Ş":
                formation = MorphotacticEngine.resolveSh(formation)
            else:
                if i == len(self.__with) - 1 and self.__with[i] == "s":
                    formation += "ş"
                else:
                    formation += self.__with[i]
            formation_to_check = formation
            i = i + 1
        return formation

    def __str__(self):
        """
        An overridden toString method which returns the with variable.

        RETURNS
        -------
        str
            With variable.
        """
        return self.__with

    def withName(self) -> str:
        """
        The withName method returns the withName variable.

        RETURNS
        -------
        str
            The withName variable.
        """
        return self.__with_name



from Dictionary.Word import Word
from MorphologicalAnalysis.MorphologicalParse import MorphologicalParse


class DisambiguatedWord(Word):

    __parse: MorphologicalParse

    def __init__(self,
                 name: str,
                 parse: MorphologicalParse):
        """
        The constructor of {@link DisambiguatedWord} class which takes a str and a MorphologicalParse
        as inputs. It creates a new MorphologicalParse with given MorphologicalParse. It generates a new instance with
        given str.
         *
         * @param name  Instances that will be a DisambiguatedWord.
         * @param parse {@link MorphologicalParse} of the {@link DisambiguatedWord}.
        """
        super().__init__(name)
        self.__parse = parse

    def getParse(self) -> MorphologicalParse:
        """
        Accessor for the MorphologicalParse.

        RETURNS
        -------
        MorphologicalParse
            MorphologicalParse.
        """
        return self.__parse

    def __repr__(self):
        return f"{self.name} {self.__parse}"



from Dictionary.Word import Word

from MorphologicalAnalysis.FsmParse import FsmParse


class FsmParseList:
    __fsm_parses: list
    longest_root_exceptions = [
        "acağı acak NOUN VERB", "acağım acak NOUN VERB", "acağımı acak NOUN VERB", "acağımız acak NOUN VERB",
        "acağın acak NOUN VERB",
        "acağına acak NOUN VERB", "acağını acak NOUN VERB", "acağının acak NOUN VERB", "acağınız acak NOUN VERB",
        "acağınıza acak NOUN VERB",
        "acağınızdır acak NOUN VERB", "acağınızı acak NOUN VERB", "acağınızın acak NOUN VERB", "acağız acak NOUN VERB",
        "acakları acak NOUN VERB",
        "acaklarını acak NOUN VERB", "acaksa acak NOUN VERB", "acaktır acak NOUN VERB", "ardım ar NOUN VERB",
        "arız ar NOUN VERB",
        "arken ar NOUN VERB", "arsa ar NOUN VERB", "arsak ar NOUN VERB", "arsanız ar NOUN VERB", "arsınız ar NOUN VERB",
        "eceği ecek NOUN VERB", "eceğim ecek NOUN VERB", "eceğimi ecek NOUN VERB", "eceğimiz ecek NOUN VERB",
        "eceğin ecek NOUN VERB",
        "eceğine ecek NOUN VERB", "eceğini ecek NOUN VERB", "eceğinin ecek NOUN VERB", "eceğiniz ecek NOUN VERB",
        "eceğinizdir ecek NOUN VERB",
        "eceğinize ecek NOUN VERB", "eceğinizi ecek NOUN VERB", "eceğinizin ecek NOUN VERB", "eceğiz ecek NOUN VERB",
        "ecekleri ecek NOUN VERB",
        "eceklerini ecek NOUN VERB", "ecekse ecek NOUN VERB", "ecektir ecek NOUN VERB", "erdim er NOUN VERB",
        "eriz er NOUN VERB",
        "erken er NOUN VERB", "erse er NOUN VERB", "ersek er NOUN VERB", "erseniz er NOUN VERB", "ersiniz er NOUN VERB",
        "ilen i VERB VERB", "ilene i VERB VERB", "ilin i VERB VERB", "ilince i VERB VERB", "imiz i ADJ NOUN",
        "in i ADJ NOUN", "inde i ADJ NOUN", "ine i ADJ NOUN", "ini i ADJ NOUN", "inin i ADJ NOUN",
        "ılan ı NOUN VERB", "ılana ı NOUN VERB", "ılın ı NOUN VERB", "ılınca ı NOUN VERB", "la la VERB NOUN",
        "lar la VERB NOUN", "lardan la VERB NOUN", "lardandır la VERB NOUN", "lardır la VERB NOUN", "ları la VERB NOUN",
        "larıdır la VERB NOUN", "larım la VERB NOUN", "larımda la VERB NOUN", "larımdan la VERB NOUN", "larımız la VERB NOUN",
        "larımıza la VERB NOUN",
        "larımızda la VERB NOUN", "larımızdan la VERB NOUN", "larımızdaydı la VERB NOUN", "larımızı la VERB NOUN",
        "larımızın la VERB NOUN",
        "larımızla la VERB NOUN", "ların la VERB NOUN", "larına la VERB NOUN", "larında la VERB NOUN",
        "larındaki la VERB NOUN",
        "larındakiler la VERB NOUN", "larındakilere la VERB NOUN", "larındakileri la VERB NOUN",
        "larındakilerin la VERB NOUN", "larından la VERB NOUN",
        "larındandır la VERB NOUN", "larındaysa la VERB NOUN", "larını la VERB NOUN", "larının la VERB NOUN",
        "larınız la VERB NOUN",
        "larınıza la VERB NOUN", "larınızda la VERB NOUN", "larınızdaki la VERB NOUN", "larınızdan la VERB NOUN",
        "larınızı la VERB NOUN",
        "larınızın la VERB NOUN", "larınızla la VERB NOUN", "larıyla la VERB NOUN", "le le VERB NOUN",
        "ler le VERB NOUN",
        "lerden le VERB NOUN", "lerdendir le VERB NOUN", "lerdir le VERB NOUN", "leri le VERB NOUN",
        "leridir le VERB NOUN",
        "lerim le VERB NOUN", "lerimden le VERB NOUN", "lerimiz le VERB NOUN", "lerimizde le VERB NOUN",
        "lerimizden le VERB NOUN",
        "lerimizdeydi le VERB NOUN", "lerimize le VERB NOUN", "lerimizi le VERB NOUN", "lerimizin le VERB NOUN",
        "lerimizle le VERB NOUN",
        "lerin le VERB NOUN", "lerinde le VERB NOUN", "lerindeki le VERB NOUN", "lerindekiler le VERB NOUN",
        "lerindekilere le VERB NOUN",
        "lerindekileri le VERB NOUN", "lerindekilerin le VERB NOUN", "lerinden le VERB NOUN",
        "lerindendir le VERB NOUN", "lerindeyse le VERB NOUN",
        "lerine le VERB NOUN", "lerini le VERB NOUN", "lerinin le VERB NOUN", "leriniz le VERB NOUN",
        "lerinizde le VERB NOUN",
        "lerinizdeki le VERB NOUN", "lerinizden le VERB NOUN", "lerinize le VERB NOUN", "lerinizi le VERB NOUN",
        "lerinizin le VERB NOUN",
        "lerinizle le VERB NOUN", "leriyle le VERB NOUN", "madan ma NOUN VERB", "malı ma NOUN VERB",
        "malıdır ma NOUN VERB", "malıdırlar ma NOUN VERB", "malılar ma NOUN VERB", "malısınız ma NOUN VERB",
        "malıyım ma NOUN VERB",
        "malıyız ma NOUN VERB", "mam ma NOUN VERB", "mama ma NOUN VERB", "mamız ma NOUN VERB", "mamıza ma NOUN VERB",
        "mamızı ma NOUN VERB", "manız ma NOUN VERB", "manızda ma NOUN VERB", "manızdır ma NOUN VERB",
        "manızı ma NOUN VERB",
        "manızla ma NOUN VERB", "ması ma NOUN VERB", "masıdır ma NOUN VERB", "masın ma NOUN VERB",
        "masına ma NOUN VERB",
        "masında ma NOUN VERB", "masındaki ma NOUN VERB", "masını ma NOUN VERB", "masıyla ma NOUN VERB",
        "mdan m NOUN NOUN",
        "meden me NOUN VERB", "meli me NOUN VERB", "melidir me NOUN VERB", "melidirler me NOUN VERB",
        "meliler me NOUN VERB",
        "melisiniz me NOUN VERB", "meliyim me NOUN VERB", "meliyiz me NOUN VERB", "mem me NOUN VERB",
        "meme me NOUN VERB",
        "memiz me NOUN VERB", "memize me NOUN VERB", "memizi me NOUN VERB", "meniz me NOUN VERB",
        "menizde me NOUN VERB",
        "menizdir me NOUN VERB", "menizi me NOUN VERB", "menizle me NOUN VERB", "mesi me NOUN VERB",
        "mesidir me NOUN VERB",
        "mesin me NOUN VERB", "mesinde me NOUN VERB", "mesindeki me NOUN VERB", "mesine me NOUN VERB",
        "mesini me NOUN VERB",
        "mesiyle me NOUN VERB", "mişse miş NOUN VERB", "mını m NOUN NOUN", "mışsa mış NOUN VERB", "mız m NOUN NOUN",
        "na n NOUN NOUN", "ne n NOUN NOUN", "nin n NOUN NOUN", "niz n NOUN NOUN",
        "nın n NOUN NOUN", "nız n NOUN NOUN", "rdim r NOUN VERB", "rdım r NOUN VERB", "riz r NOUN VERB",
        "rız r NOUN VERB", "rken r NOUN VERB", "rken r NOUN VERB", "rsa r NOUN VERB", "rsak r NOUN VERB",
        "rsanız r NOUN VERB", "rse r NOUN VERB", "rsek r NOUN VERB", "rseniz r NOUN VERB", "rsiniz r NOUN VERB",
        "rsınız r NOUN VERB", "sa sa VERB ADJ", "se se VERB ADJ", "ulan u NOUN VERB", "un un VERB NOUN",
        "üne ün VERB NOUN", "unun un VERB NOUN", "ince i NOUN VERB", "unca u NOUN VERB", "ınca ı NOUN VERB",
        "unca un NOUN VERB", "ilen ile VERB VERB"]

    def __init__(self, fsmParses: list):
        """
        A constructor of FsmParseList class which takes a list fsmParses as an input. First it sorts
        the items of the list then loops through it, if the current item's transitions equal to the next item's
        transitions, it removes the latter item. At the end, it assigns this list to the fsmParses variable.

        PARAMETERS
        ----------
        fsmParses : list
            FsmParse list input.
        """
        fsmParses.sort()
        i = 0
        while i < len(fsmParses) - 1:
            if fsmParses[i].transitionList() == fsmParses[i + 1].transitionList():
                fsmParses.pop(i + 1)
                i = i - 1
            i = i + 1
        self.__fsm_parses = fsmParses

    def size(self) -> int:
        """
        The size method returns the size of fsmParses list.

        RETURNS
        -------
        int
            The size of fsmParses list.
        """
        return len(self.__fsm_parses)

    def getFsmParse(self, index: int) -> FsmParse:
        """
        The getFsmParse method takes an integer index as an input and returns the item of fsmParses list at given index.

        PARAMETERS
        ----------
        index : int
            Integer input.

        RETURNS
        -------
        FsmParse
            The item of fsmParses list at given index.
        """
        return self.__fsm_parses[index]

    def rootWords(self) -> str:
        """
        The rootWords method gets the first item's root of fsmParses list and uses it as currentRoot. Then loops through
        the fsmParses, if the current item's root does not equal to the currentRoot, it then assigns it as the
        currentRoot and accumulates root words in a string result.

        RETURNS
        -------
        str
            String result that has root words.
        """
        result = self.__fsm_parses[0].getWord().getName()
        current_root = result
        for i in range(1, len(self.__fsm_parses)):
            if self.__fsm_parses[i].getWord().getName() != current_root:
                current_root = self.__fsm_parses[i].getWord().getName()
                result = result + "$" + current_root
        return result

    def reduceToParsesWithSameRootAndPos(self, currentWithPos: Word):
        """
        The reduceToParsesWithSameRootAndPos method takes a Word currentWithPos as an input and loops i times till
        i equals to the size of the fsmParses list. If the given currentWithPos does not equal to the ith item's
        root and the MorphologicalTag of the first inflectional of fsmParses, it removes the ith item from the list.

        PARAMETERS
        ----------
        currentWithPos : Word
            Word input.
        """
        i = 0
        while i < len(self.__fsm_parses):
            if self.__fsm_parses[i].getWordWithPos() != currentWithPos:
                self.__fsm_parses.pop(i)
            else:
                i = i + 1

    def getParseWithLongestRootWord(self) -> FsmParse:
        """
        The getParseWithLongestRootWord method returns the parse with the longest root word. If more than one parse has
        the longest root word, the first parse with that root is returned. If the longest root word belongs to an
        exceptional case, the parse with the next longest root word that does not, is returned.

        RETURNS
        -------
        Parse with the longest root word.
        """
        max_length = -1
        best_parse = None
        if len(self.__fsm_parses) > 0:
            best_parse = self.__fsm_parses[0]
        for current_parse in self.__fsm_parses:
            if len(current_parse.getWord().getName()) > max_length and not self.isLongestRootException(current_parse):
                max_length = len(current_parse.getWord().getName())
                best_parse = current_parse
        return best_parse

    def isLongestRootException(self, fsmParse: FsmParse) -> bool:
        """
        The isLongestRootException method returns true if the longest root word belongs to an exceptional case, false
        otherwise.
        :param fsmParse: FsmParse input
        :return: true if the longest root belongs to an exceptional case, false otherwise.
        """
        surface_form = fsmParse.getSurfaceForm()
        root = fsmParse.getWord().getName()
        for longest_root_exception in self.longest_root_exceptions:
            exception_items = longest_root_exception.split(" ")
            surface_form_ending = exception_items[0]
            longest_root_ending = exception_items[1]
            longest_root_pos = exception_items[2]
            possible_root_pos = exception_items[3]
            possible_root = surface_form.replace(surface_form_ending, "")
            if surface_form.endswith(surface_form_ending) and root.endswith(longest_root_ending) \
                    and fsmParse.getRootPos() == longest_root_pos:
                for current_parse in self.__fsm_parses:
                    if current_parse.getWord().getName() == possible_root \
                            and current_parse.getRootPos() == possible_root_pos:
                        return True
        return False

    def reduceToParsesWithSameRoot(self, currentRoot: str):
        """
        The reduceToParsesWithSameRoot method takes a str currentWithPos as an input and loops i times till
        i equals to the size of the fsmParses list. If the given currentRoot does not equal to the root of ith item of
        fsmParses, it removes the ith item from the list.

        PARAMETERS
        ----------
        currentRoot : str
            String input.
        """
        i = 0
        while i < len(self.__fsm_parses):
            if self.__fsm_parses[i].getWord().getName() != currentRoot:
                self.__fsm_parses.pop(i)
            else:
                i = i + 1

    def constructParseListForDifferentRootWithPos(self) -> list:
        """
        The constructParseListForDifferentRootWithPos method initially creates a result list then loops through the
        fsmParses list. For the first iteration, it creates new list as initial, then adds the
        first item od fsmParses to initial and also add this initial list to the result list. For the following
        iterations,
        it checks whether the current item's root with the MorphologicalTag of the first inflectional
        equal to the previous item's  root with the MorphologicalTag of the first inflectional. If so, it adds that item
        to the result list, if not it creates new list as initial and adds the first item od fsmParses
        to initial and also add this initial list to the result list.

        RETURNS
        -------
        result : list
            list type of FsmParseList.
        """
        result = []
        i = 0
        while i < len(self.__fsm_parses):
            if i == 0:
                initial = [self.__fsm_parses[i]]
                result.append(FsmParseList(initial))
            else:
                if self.__fsm_parses[i].getWordWithPos() == self.__fsm_parses[i - 1].getWordWithPos():
                    result[len(result) - 1].__fsm_parses.append(self.__fsm_parses[i])
                else:
                    initial = [self.__fsm_parses[i]]
                    result.append(FsmParseList(initial))
            i = i + 1
        return result

    def parsesWithoutPrefixAndSuffix(self) -> str:
        """
        The parsesWithoutPrefixAndSuffix method first creates a str array named analyses with the size of fsmParses
        list's size.

        If the size is just 1, it then returns the first item's transitionList, if it is greater than 1, loops through
        the fsmParses and puts the transitionList of each item to the analyses array.

        If the removePrefix condition holds, it loops through the analyses array and takes each item's substring after
        the first + sign and updates that item of analyses array with that substring.

        If the removeSuffix condition holds, it loops through the analyses array and takes each item's substring till
        the last + sign and updates that item of analyses array with that substring.

        It then removes the duplicate items of analyses array and returns a result str that has the accumulated items of
        analyses array.

        RETURNS
        -------
        str
            result str that has the accumulated items of analyses array.
        """
        analyses = [""] * len(self.__fsm_parses)
        remove_prefix = True
        remove_suffix = True
        if len(self.__fsm_parses) == 1:
            return self.__fsm_parses[0].transitionList()[self.__fsm_parses[0].transitionList().index("+") + 1]
        for i in range(len(self.__fsm_parses)):
            analyses[i] = self.__fsm_parses[i].transitionList()
        while remove_prefix:
            remove_prefix = True
            for i in range(len(self.__fsm_parses) - 1):
                if "+" not in analyses[i] or "+" not in analyses[i + 1] or \
                        analyses[i][:analyses[i].index("+") + 1] != analyses[i + 1][:analyses[i + 1].index("+") + 1]:
                    remove_prefix = False
                    break
            if remove_prefix:
                for i in range(len(self.__fsm_parses)):
                    analyses[i] = analyses[i][analyses[i].index("+") + 1:]
        while remove_suffix:
            remove_suffix = True
            for i in range(len(self.__fsm_parses) - 1):
                if "+" not in analyses[i] or "+" not in analyses[i + 1] or \
                        analyses[i][analyses[i].rindex("+") + 1] != analyses[i + 1][analyses[i + 1].rindex("+") + 1]:
                    remove_suffix = False
                    break
            if remove_suffix:
                for i in range(len(self.__fsm_parses)):
                    analyses[i] = analyses[i][:analyses[i].rindex("+")]
        for i in range(len(analyses)):
            for j in range(i + 1, len(analyses)):
                if analyses[i] > analyses[j]:
                    analyses[i], analyses[j] = analyses[j], analyses[i]
        result = analyses[0]
        for i in range(1, len(analyses)):
            result = result + "$" + analyses[i]
        return result

    def __str__(self) -> str:
        """
        The overridden toString method loops through the fsmParses list and accumulates the items to a result str.

        RETURNS
        -------
        str
            Result str that has the items of fsmParses list.
        """
        result = ""
        for i in range(len(self.__fsm_parses)):
            result += self.__fsm_parses[i] + "\n"
        return result



from __future__ import annotations

from Dictionary.TxtWord import TxtWord
from Dictionary.Word import Word

from MorphologicalAnalysis.InflectionalGroup import InflectionalGroup
from MorphologicalAnalysis.MorphologicalParse import MorphologicalParse
from MorphologicalAnalysis.State import State


class FsmParse(MorphologicalParse):
    __suffix_list: list
    __form_list: list
    __transition_list: list
    __with_list: list
    __initial_pos: str
    __pos: str
    __form: str
    __verb_agreement: str
    __possesive_agreement: str

    def __init__(self,
                 root: Word | str | int | float,
                 startState=None):
        """
        Another constructor of FsmParse class which takes a TxtWord root and a State as inputs.
        First, initializes root variable with this TxtWord. It also initializes form with root's name, pos and
        initialPos with given State's POS, creates 4 new list suffixList, formList, transitionList
        and withList and adds given State to suffixList, form to formList.

        PARAMETERS
        ----------
        root : Word
            Word input.
        startState : State
            State input.
        """
        if isinstance(root, Word):
            self.root = root
        elif isinstance(root, int):
            self.root = TxtWord(str(root))
            self.root.addFlag("IS_SAYI")
        elif isinstance(root, float):
            self.root = TxtWord(str(root))
            self.root.addFlag("IS_SAYI")
        elif isinstance(root, str):
            self.root = TxtWord(root)
        if startState is not None:
            self.__form = self.root.getName()
            self.__pos = startState.getPos()
            self.__initial_pos = startState.getPos()
            self.__suffix_list = []
            self.__suffix_list.append(startState)
            self.__form_list = []
            self.__form_list.append(self.__form)
            self.__transition_list = []
            self.__with_list = []
        self.__verb_agreement = None
        self.__possesive_agreement = None

    def __eq__(self, other):
        return self.transitionList() == other.transitionList()

    def __lt__(self, other):
        return self.transitionList() < other.transitionList()

    def __gt__(self, other):
        return self.transitionList() > other.transitionList()

    def constructInflectionalGroups(self):
        """
        The constructInflectionalGroups method initially calls the transitionList method and assigns the resulting str
        to the parse variable and creates a new list as iGs. If parse str contains a derivational boundary
        it adds the substring starting from the 0 to the index of derivational boundary to the iGs. If it does not
        contain a DB, it directly adds parse to the iGs. Then, creates and initializes new list as inflectionalGroups
        and fills with the items of iGs.
        """
        parse = self.transitionList()
        iGs = []
        while "^DB+" in parse:
            iGs.append(parse[:parse.index("^DB+")])
            parse = parse[parse.index("^DB+") + 4:]
        iGs.append(parse)
        self.inflectional_groups = []
        self.inflectional_groups.append(InflectionalGroup(iGs[0][iGs[0].index("+") + 1:]))
        for i in range(1, len(iGs)):
            self.inflectional_groups.append(InflectionalGroup(iGs[i]))

    def getVerbAgreement(self) -> str:
        """
        Getter for the verbAgreement variable.

        RETURNS
        -------
        str
            The verbAgreement variable.
        """
        return self.__verb_agreement

    def getPossesiveAgreement(self) -> str:
        """
        Getter for the possesiveAgreement variable.

        RETURNS
        -------
        str
            The possesiveAgreement variable.
        """
        return self.__possesive_agreement

    def setAgreement(self, transitionName: str):
        """
        The setAgreement method takes a str transitionName as an input and if it is one of the A1SG, A2SG, A3SG,
        A1PL, A2PL or A3PL it assigns transitionName input to the verbAgreement variable. Or if it is ine of the PNON,
        P1SG, P2SG,P3SG, P1PL, P2PL or P3PL it assigns transitionName input to the possesiveAgreement variable.

        PARAMETERS
        ----------
        transitionName : str
            String input.
        """
        if transitionName == "A1SG" or transitionName == "A2SG" or transitionName == "A3SG" or transitionName == "A1PL"\
                or transitionName == "A2PL" or transitionName == "A3PL":
            self.__verb_agreement = transitionName
        if transitionName == "PNON" or transitionName == "P1SG" or transitionName == "P2SG" or transitionName == "P3SG"\
                or transitionName == "P1PL" or transitionName == "P2PL" or transitionName == "P3PL":
            self.__possesive_agreement = transitionName

    def getLastLemmaWithTag(self, pos: str) -> str:
        """
        The getLastLemmaWithTag method takes a String input pos as an input. If given pos is an initial pos then it
        assigns root to the lemma, and assign null otherwise. Then, it loops i times where i ranges from 1 to size of
        the formList, if the item at i-1 of transitionList is not null and contains a derivational boundary with pos but
        not with ZERO, it assigns the ith item of formList to lemma.

        PARAMETERS
        ----------
        pos : str
            String input.

        RETURNS
        -------
        str
            String output lemma.
        """
        if self.__initial_pos is not None and self.__initial_pos == pos:
            lemma = self.root.getName()
        else:
            lemma = None
        for i in range(1, len(self.__form_list)):
            if self.__transition_list[i - 1] is not None and ("^DB+" + pos) in self.__transition_list[i - 1] and \
                    ("^DB+" + pos + "+ZERO") not in self.__transition_list[i - 1]:
                lemma = self.__form_list[i]
        return lemma

    def getLastLemma(self) -> str:
        """
        The getLastLemma method initially assigns root as lemma. Then, it loops i times where i ranges from 1 to size of
        the formList, if the item at i-1 of transitionList is not null and contains a derivational boundary, it assigns
        the ith item of formList to lemma.

        RETURNS
        -------
        str
            String output lemma.
        """
        lemma = self.root.getName()
        for i in range(1, len(self.__form_list)):
            if self.__transition_list[i - 1] is not None and "^DB+" in self.__transition_list[i - 1]:
                lemma = self.__form_list[i]
        return lemma

    def addSuffix(self,
                  suffix: State,
                  form: str,
                  transition: str,
                  withName: str,
                  toPos: str):
        """
        The addSuffix method takes 5 different inputs; State suffix, str form, transition, with and toPos.
        If the pos of given input suffix is not None, it then assigns it to the pos variable. If the pos of the given
        suffix is None but given toPos is not None than it assigns toPos to pos variable. At the end, it adds suffix to
        the suffixList, form to the formList, transition to the transitionList and if given with is not 0, it is also
        added to withList.

        PARAMETERS
        ----------
        suffix : State
            State input.
        form : str
            String input.
        transition : str
            String input.
        withName : str
            String input.
        toPos : str
            String input.
        """
        if suffix.getPos() is not None:
            self.__pos = suffix.getPos()
        else:
            if toPos is not None:
                self.__pos = toPos
        self.__suffix_list.append(suffix)
        self.__form_list.append(form)
        self.__transition_list.append(transition)
        if withName != "0":
            self.__with_list.append(withName)
        self.__form = form

    def getSurfaceForm(self) -> str:
        """
        Getter for the form variable.

        RETURNS
        -------
        str
            The form variable.
        """
        return self.__form

    def getStartState(self) -> State:
        """
        The getStartState method returns the first item of suffixList list.

        RETURNS
        -------
        State
            The first item of suffixList list.
        """
        return self.__suffix_list[0]

    def getFinalPos(self) -> str:
        """
        Getter for the pos variable.

        RETURNS
        -------
        str
            The pos variable.
        """
        return self.__pos

    def getInitialPos(self) -> str:
        """
        Getter for the initialPos variable.

        RETURNS
        -------
        str
            The initialPos variable.
        """
        return self.__initial_pos

    def setForm(self, name: str):
        """
        The setForm method takes a str name as an input and assigns it to the form variable, then it removes the first
        item of formList list and adds the given name to the formList.

        PARAMETERS
        ----------
        name : str
            String input to set form.
        """
        self.__form = name
        self.__form_list.pop(0)
        self.__form_list.append(name)

    def getFinalSuffix(self) -> State:
        """
        The getFinalSuffix method returns the last item of suffixList list.

        RETURNS
        -------
        State
            The last item of suffixList list.
        """
        return self.__suffix_list[len(self.__suffix_list) - 1]

    def headerTransition(self) -> str:
        """
        The headerTransition method gets the first item of formList and checks for cases;

        If it is <DOC>, it returns "<DOC>+BDTAG" which indicates the beginning of a document.
        If it is </DOC>, it returns "</DOC>+EDTAG" which indicates the ending of a document.
        If it is <TITLE>, it returns "<TITLE>+BTTAG" which indicates the beginning of a title.
        If it is </TITLE>, it returns "</TITLE>+ETTAG" which indicates the ending of a title.
        If it is "<S>", it returns "<S>+BSTAG" which indicates the beginning of a sentence.
        If it is "</S>, it returns "</S>+ESTAG" which indicates the ending of a sentence.

        RETURNS
        -------
        str
            Corresponding tags of the headers and an empty {@link String} if any case does not match.
        """
        if self.__form_list[0] == "<DOC>":
            return "<DOC>+BDTAG"
        elif self.__form_list[0] == "</DOC>":
            return "</DOC>+EDTAG"
        elif self.__form_list[0] == "<TITLE>":
            return "<TITLE>+BTTAG"
        elif self.__form_list[0] == "</TITLE>":
            return "</TITLE>+ETTAG"
        elif self.__form_list[0] == "<S>":
            return "<S>+BSTAG"
        elif self.__form_list[0] == "</S>":
            return "</S>+ESTAG"
        else:
            return ""

    def pronounTransition(self) -> str:
        """
        The pronounTransition method gets the first item of formList and checks for cases;

        If it is "kendi", it returns kendi+PRON+REFLEXP which indicates a reflexive pronoun.
        If it is one of the "hep, öbür, topu, öteki, kimse, hiçbiri, tümü, çoğu, hepsi, herkes, başkası, birçoğu,
        birçokları, biri, birbirleri, birbiri, birkaçı, böylesi, diğeri, cümlesi, bazı, kimi", it returns
        +PRON+QUANTP which indicates a quantitative pronoun.
        If it is one of the "o, bu, şu" and if it is "o" it also checks the first item of suffixList and if it is a
        PronounRoot(DEMONS), it returns +PRON+DEMONSP which indicates a demonstrative pronoun.
        If it is "ben", it returns +PRON+PERS+A1SG+PNON which indicates a 1st person singular agreement.
        If it is "sen", it returns +PRON+PERS+A2SG+PNON which indicates a 2nd person singular agreement.
        If it is "o" and the first item of suffixList, if it is a PronounRoot(PERS), it returns +PRON+PERS+A3SG+PNON
        which indicates a 3rd person singular agreement.
        If it is "biz", it returns +PRON+PERS+A1PL+PNON which indicates a 1st person plural agreement.
        If it is "siz", it returns +PRON+PERS+A2PL+PNON which indicates a 2nd person plural agreement.
        If it is "onlar" and the first item of suffixList, if it is a PronounRoot(PERS), it returns
        o+PRON+PERS+A3PL+PNON which indicates a 3rd person plural agreement.
        If it is one of the "nere, ne, kim, hangi", it returns +PRON+QUESP which indicates a question pronoun.

        RETURNS
        -------
        str
            Corresponding transitions of pronouns and an empty str if any case does not match.
        """
        if self.__form_list[0] == "kendi":
            return "kendi+PRON+REFLEXP"
        elif self.__form_list[0] == "hep" or self.__form_list[0] == "öbür" or self.__form_list[0] == "topu" \
                or self.__form_list[0] == "öteki" or self.__form_list[0] == "kimse" or self.__form_list[0] == "hiçbiri" \
                or self.__form_list[0] == "tümü" or self.__form_list[0] == "çoğu" or self.__form_list[0] == "hepsi" \
                or self.__form_list[0] == "herkes" or self.__form_list[0] == "başkası" or self.__form_list[0] == "birçoğu"\
                or self.__form_list[0] == "birçokları" or self.__form_list[0] == "birbiri" \
                or self.__form_list[0] == "birbirleri" or self.__form_list[0] == "biri" \
                or self.__form_list[0] == "birkaçı" or self.__form_list[0] == "böylesi" or self.__form_list[0] == "diğeri"\
                or self.__form_list[0] == "cümlesi" or self.__form_list[0] == "bazı" or self.__form_list[0] == "kimi":
            return self.__form_list[0] + "+PRON+QUANTP"
        elif (self.__form_list[0] == "o" and self.__suffix_list[0].getName() == "PronounRoot(DEMONS)") \
                or self.__form_list[0] == "bu" or self.__form_list[0] == "şu":
            return self.__form_list[0] + "+PRON+DEMONSP"
        elif self.__form_list[0] == "ben":
            return self.__form_list[0] + "+PRON+PERS+A1SG+PNON"
        elif self.__form_list[0] == "sen":
            return self.__form_list[0] + "+PRON+PERS+A2SG+PNON"
        elif self.__form_list[0] == "o" and self.__suffix_list[0].getName() == "PronounRoot(PERS)":
            return self.__form_list[0] + "+PRON+PERS+A3SG+PNON"
        elif self.__form_list[0] == "biz":
            return self.__form_list[0] + "+PRON+PERS+A1PL+PNON"
        elif self.__form_list[0] == "siz":
            return self.__form_list[0] + "+PRON+PERS+A2PL+PNON"
        elif self.__form_list[0] == "onlar":
            return "o+PRON+PERS+A3PL+PNON"
        elif self.__form_list[0] == "nere" or self.__form_list[0] == "ne" or self.__form_list[0] == "kaçı" \
                or self.__form_list[0] == "kim" or self.__form_list[0] == "hangi":
            return self.__form_list[0] + "+PRON+QUESP"
        else:
            return ""

    def transitionList(self) -> str:
        """
        The transitionList method first creates an empty {@link String} result, then gets the first item of suffixList
        and checks for cases; If it is one of the "NominalRoot, NominalRootNoPossesive, CompoundNounRoot,
        NominalRootPlural", it assigns concatenation of first item of formList and +NOUN to the result String.
        Ex : Birincilik

        If it is one of the "VerbalRoot, PassiveHn", it assigns concatenation of first item of formList and +VERB to the
        result String.
        Ex : Başkalaştı

        If it is "CardinalRoot", it assigns concatenation of first item of formList and +NUM+CARD to the result String.
        Ex : Onuncu

        If it is "FractionRoot", it assigns concatenation of first item of formList and NUM+FRACTION to the result
        String.
        Ex : 1/2

        If it is "TimeRoot", it assigns concatenation of first item of formList and +TIME to the result String.
        Ex : 14:28

        If it is "RealRoot", it assigns concatenation of first item of formList and +NUM+REAL to the result String.
        Ex : 1.2

        If it is "Punctuation", it assigns concatenation of first item of formList and +PUNC to the result String.
        Ex : ,

        If it is "Hashtag", it assigns concatenation of first item of formList and +HASHTAG to the result String.
        Ex : #

        If it is "DateRoot", it assigns concatenation of first item of formList and +DATE to the result String.
        Ex : 11/06/2018

        If it is "RangeRoot", it assigns concatenation of first item of formList and +RANGE to the result String.
        Ex : 3-5

        If it is "Email", it assigns concatenation of first item of formList and +EMAIL to the result String.
        Ex : abc@

        If it is "PercentRoot", it assigns concatenation of first item of formList and +PERCENT to the result String.
        Ex : %12.5

        If it is "DeterminerRoot", it assigns concatenation of first item of formList and +DET to the result String.
        Ex : Birtakım

        If it is "ConjunctionRoot", it assigns concatenation of first item of formList and +CONJ to the result String.
        Ex : Ama

        If it is "AdverbRoot", it assigns concatenation of first item of formList and +ADV to the result String.
        Ex : Acilen

        If it is "ProperRoot", it assigns concatenation of first item of formList and +NOUN+PROP to the result String.
        Ex : Ahmet

        If it is "HeaderRoot", it assigns the result of the headerTransition method to the result String.
        Ex : &lt;DOC&gt;

        If it is "InterjectionRoot", it assigns concatenation of first item of formList and +INTERJ to the result
        String.
        Ex : Hey

        If it is "DuplicateRoot", it assigns concatenation of first item of formList and +DUP to the result String.
        Ex : Allak

        If it is "CodeRoot", it assigns concatenation of first item of formList and +CODE to the result String.
        Ex : 5000-WX

        If it is "MetricRoot", it assigns concatenation of first item of formList and +METRIC to the result String.
        Ex : 6cmx12cm

        If it is "QuestionRoot", it assigns concatenation of first item of formList and +QUES to the result String.
        Ex : Mı

        If it is "PostP", and the first item of formList is one of the "karşı, ilişkin, göre, kadar, ait, yönelik,
        rağmen, değin, dek, doğru, karşın, dair, atfen, binaen, hitaben, istinaden, mahsuben, mukabil, nazaran", it
        assigns concatenation of first item of formList and +POSTP+PCDAT to the result String.
        Ex : İlişkin

        If it is "PostP", and the first item of formList is one of the "sonra, önce, beri, fazla, dolayı, itibaren,
        başka, çok, evvel, ötürü, yana, öte, aşağı, yukarı, dışarı, az, gayrı", it assigns concatenation of first
        item of formList and +POSTP+PCABL to the result String.
        Ex : Başka

        If it is "PostP", and the first item of formList is "yanısıra", it assigns concatenation of first
        item of formList and +POSTP+PCGEN to the result String.
        Ex : Yanısıra

        If it is "PostP", and the first item of formList is one of the "birlikte, beraber", it assigns concatenation of
        first item of formList and +PPOSTP+PCINS to the result String.
        Ex : Birlikte

        If it is "PostP", and the first item of formList is one of the "aşkın, takiben", it assigns concatenation of
        first item of formList and +POSTP+PCACC to the result String.
        Ex : Takiben

        If it is "PostP", it assigns concatenation of first item of formList and +POSTP+PCNOM to the result String.

        If it is "PronounRoot", it assigns result of the pronounTransition method to the result String.
        Ex : Ben

        If it is "OrdinalRoot", it assigns concatenation of first item of formList and +NUM+ORD to the result String.
        Ex : Altıncı

        If it starts with "Adjective", it assigns concatenation of first item of formList and +ADJ to the result String.
        Ex : Güzel

        At the end, it loops through the formList and concatenates each item with result {@link String}.

        RETURNS
        -------
        str
            str result accumulated with items of formList.
        """
        result = ""
        if self.__suffix_list[0].getName() == "NominalRoot" \
                or self.__suffix_list[0].getName() == "NominalRootNoPossesive" \
                or self.__suffix_list[0].getName() == "CompoundNounRoot" \
                or self.__suffix_list[0].getName() == "NominalRootPlural":
            result = self.__form_list[0] + "+NOUN"
        elif self.__suffix_list[0].getName().startswith("VerbalRoot") or self.__suffix_list[0].getName() == "PassiveHn":
            result = self.__form_list[0] + "+VERB"
        elif self.__suffix_list[0].getName() == "CardinalRoot":
            result = self.__form_list[0] + "+NUM+CARD"
        elif self.__suffix_list[0].getName() == "FractionRoot":
            result = self.__form_list[0] + "+NUM+FRACTION"
        elif self.__suffix_list[0].getName() == "TimeRoot":
            result = self.__form_list[0] + "+TIME"
        elif self.__suffix_list[0].getName() == "RealRoot":
            result = self.__form_list[0] + "+NUM+REAL"
        elif self.__suffix_list[0].getName() == "Punctuation":
            result = self.__form_list[0] + "+PUNC"
        elif self.__suffix_list[0].getName() == "Hashtag":
            result = self.__form_list[0] + "+HASHTAG"
        elif self.__suffix_list[0].getName() == "DateRoot":
            result = self.__form_list[0] + "+DATE"
        elif self.__suffix_list[0].getName() == "RangeRoot":
            result = self.__form_list[0] + "+RANGE"
        elif self.__suffix_list[0].getName() == "Email":
            result = self.__form_list[0] + "+EMAIL"
        elif self.__suffix_list[0].getName() == "PercentRoot":
            result = self.__form_list[0] + "+PERCENT"
        elif self.__suffix_list[0].getName() == "DeterminerRoot":
            result = self.__form_list[0] + "+DET"
        elif self.__suffix_list[0].getName() == "ConjunctionRoot":
            result = self.__form_list[0] + "+CONJ"
        elif self.__suffix_list[0].getName() == "AdverbRoot":
            result = self.__form_list[0] + "+ADV"
        elif self.__suffix_list[0].getName() == "ProperRoot":
            result = self.__form_list[0] + "+NOUN+PROP"
        elif self.__suffix_list[0].getName() == "HeaderRoot":
            result = self.headerTransition()
        elif self.__suffix_list[0].getName() == "InterjectionRoot":
            result = self.__form_list[0] + "+INTERJ"
        elif self.__suffix_list[0].getName() == "DuplicateRoot":
            result = self.__form_list[0] + "+DUP"
        elif self.__suffix_list[0].getName() == "CodeRoot":
            result = self.__form_list[0] + "+CODE"
        elif self.__suffix_list[0].getName() == "MetricRoot":
            result = self.__form_list[0] + "+METRIC"
        elif self.__suffix_list[0].getName() == "QuestionRoot":
            result = "mi+QUES"
        elif self.__suffix_list[0].getName() == "PostP":
            if self.__form_list[0] == "karşı" or self.__form_list[0] == "ilişkin" or self.__form_list[0] == "göre" \
                    or self.__form_list[0] == "kadar" or self.__form_list[0] == "ait" or self.__form_list[0] == "yönelik" \
                    or self.__form_list[0] == "rağmen" or self.__form_list[0] == "değin" or self.__form_list[0] == "dek" \
                    or self.__form_list[0] == "doğru" or self.__form_list[0] == "karşın" or self.__form_list[0] == "dair" \
                    or self.__form_list[0] == "atfen" or self.__form_list[0] == "binaen" or self.__form_list[
                0] == "hitaben" \
                    or self.__form_list[0] == "istinaden" or self.__form_list[0] == "mahsuben" \
                    or self.__form_list[0] == "mukabil" or self.__form_list[0] == "nazaran":
                result = self.__form_list[0] + "+POSTP+PCDAT"
            elif self.__form_list[0] == "sonra" or self.__form_list[0] == "önce" or self.__form_list[0] == "beri" \
                    or self.__form_list[0] == "fazla" or self.__form_list[0] == "dolayı" or self.__form_list[
                0] == "itibaren" \
                    or self.__form_list[0] == "başka" or self.__form_list[0] == "çok" or self.__form_list[0] == "evvel" \
                    or self.__form_list[0] == "ötürü" or self.__form_list[0] == "yana" or self.__form_list[0] == "öte" \
                    or self.__form_list[0] == "aşağı" or self.__form_list[0] == "yukarı" \
                    or self.__form_list[0] == "dışarı" or self.__form_list[0] == "az" or self.__form_list[0] == "gayrı":
                result = self.__form_list[0] + "+POSTP+PCABL"
            elif self.__form_list[0] == "yanısıra":
                result = self.__form_list[0] + "+POSTP+PCGEN"
            elif self.__form_list[0] == "birlikte" or self.__form_list[0] == "beraber":
                result = self.__form_list[0] + "+POSTP+PCINS"
            elif self.__form_list[0] == "aşkın" or self.__form_list[0] == "takiben":
                result = self.__form_list[0] + "+POSTP+PCACC"
            else:
                result = self.__form_list[0] + "+POSTP+PCNOM"
        elif self.__suffix_list[0].getName().startswith("PronounRoot"):
            result = self.pronounTransition()
        elif self.__suffix_list[0].getName() == "OrdinalRoot":
            result = self.__form_list[0] + "+NUM+ORD"
        elif self.__suffix_list[0].getName().startswith("Adjective"):
            result = self.__form_list[0] + "+ADJ"
        for transition in self.__transition_list:
            if transition is not None:
                if not transition.startswith("^"):
                    result = result + "+" + transition
                else:
                    result = result + transition
        return result

    def suffixList(self) -> str:
        """
        The suffixList method gets the first items of suffixList and formList and concatenates them with parenthesis and
        assigns a String result. Then, loops through the formList and it the current ith item is not equal to previous
        item it accumulates ith items of formList and suffixList to the result str.

        RETURNS
        -------
        str
            result str accumulated with the items of formList and suffixList.
        """
        result = self.__suffix_list[0].getName() + "(" + self.__form_list[0] + ")"
        for i in range(1, len(self.__form_list)):
            if self.__form_list[i] != self.__form_list[i - 1]:
                result = result + "+" + self.__suffix_list[i].getName() + "(" + self.__form_list[i] + ")"
        return result

    def getWithList(self) -> list:
        """
        Accessor to the __with_list list.
        :return: __with_list list.
        """
        return self.__with_list

    def withList(self) -> str:
        """
        The withList method gets the root as a result str then loops through the withList and concatenates each item
        with result str.

        RETURNS
        -------
        str
            result str accumulated with items of withList.
        """
        result = self.root.getName()
        for a_with in self.__with_list:
            result = result + "+" + a_with
        return result

    def restoreOriginalForm(self, original: str, pronunciation: str):
        """
        In order to morphologically parse special proper nouns in Turkish, whose affixes obeys not the original but their
        pronunciations, the morphologicalAnalysis method replaces the original word with its pronunciation and do the
        rest. This method reverts it back, that is it restores its original form by replacing the pronunciations in the
        parses with the original form.
        :param original: Original form of the proper noun.
        :param pronunciation: Pronunciation of the proper noun.
        """
        self.root = TxtWord(original, "IS_OA")
        self.__form = original + self.__form[len(pronunciation):]
        for i in range(0, len(self.__form_list)):
            self.__form_list[i] = original + self.__form_list[i][len(pronunciation):]

    def __str__(self) -> str:
        """
        The overridden str method which returns transitionList method.

        RETURNS
        -------
        str
            Returns transitionList method.
        """
        return self.transitionList()



from Corpus.Corpus import Corpus
from Corpus.Sentence import Sentence

from DisambiguationCorpus.DisambiguatedWord import DisambiguatedWord
from MorphologicalAnalysis.MorphologicalParse import MorphologicalParse


class DisambiguationCorpus(Corpus):

    def __init__(self, fileName=None):
        """
        Constructor which creates a list of sentences and a CounterHashMap of wordList.
        """
        super().__init__()
        if fileName is not None:
            input_file = open(fileName, "r", encoding="utf8")
            lines = input_file.readlines()
            new_sentence = Sentence()
            for line in lines:
                word = line[:line.index("\t")]
                parse = line[line.index("\t") + 1:]
                if len(word) > 0 and len(parse) > 0:
                    new_word = DisambiguatedWord(word, MorphologicalParse(parse.strip()))
                    if word == "<S>":
                        new_sentence = Sentence()
                    elif word == "</S>":
                        self.addSentence(new_sentence)
                    elif word == "<DOC>" or word == "</DOC>" or word == "<TITLE>" or word == "</TITLE>":
                        pass
                    else:
                        new_sentence.addWord(new_word)
            input_file.close()

    def writeToFile(self, fileName: str):
        """
        The writeToFile method takes a str file name as an input and writes the elements of sentences list
        to this file with proper tags which indicates the beginnings and endings of the document and sentence.

        PARAMETERS
        ----------
        fileName : str
            File which will be filled with the sentences.
        """
        output_file = open(fileName, "w", encoding="utf8")
        output_file.write("<DOC>\t<DOC>+BDTag\n")
        for sentence in self.sentences:
            output_file.write("<S>\t<S>+BSTag\n")
            for word in sentence.words:
                if isinstance(word, DisambiguatedWord):
                    output_file.write(word.getName() + "\t" + word.getParse().__str__() + "\n")
            output_file.write("</S>\t</S>+ESTag\n")
        output_file.write("</DOC>\t</DOC>+EDTag")
        output_file.close()



from Dictionary.Word import Word

from MorphologicalAnalysis.MorphologicalParse import MorphologicalParse
from MorphologicalAnalysis.MorphologicalTag import MorphologicalTag


class MetamorphicParse:
    """
    MetaMorphemes that can be used.
    """
    meta_morphemes = ["Ar", "Ar", "CA", "CA",
                      "CA", "cAsHnA", "CH", "CHk",
                      "DA", "DAn", "DH", "DHk",
                      "DHkCA", "DHr", "DHr", "DHr",
                      "H", "Hl", "Hm", "Hn",
                      "Hn", "Hn", "HmHz", "HncH",
                      "HnHz", "Hr", "Hr", "Hs",
                      "Ht", "Hyor", "Hz", "k",
                      "ki", "kü", "lAn", "lAr",
                      "lArDHr", "lArH", "lArH'", "lAs",
                      "lH", "lHk", "lHm", "m",
                      "mA", "mA", "mAcA", "mAdAn",
                      "mAk", "mAksHzHn", "mAktA", "mAlH",
                      "mAzlHk", "mHs", "n", "n",
                      "nA", "ncA", "nDA", "nDAn",
                      "nH", "nHn", "nHz", "nlAr",
                      "SA", "SAl", "sH", "SH",
                      "SH", "SHn", "SHnHz", "SHnlAr",
                      "SHz", "ŞAr", "t", "yA",
                      "yA", "yAbil", "yAcAk", "yAcAk",
                      "yAdur", "yAgel", "yAlH", "yAmA",
                      "yAmAdAn", "yAn", "yArAk", "yAsH",
                      "yDH", "yH", "yHcH", "yHm",
                      "yHn", "yHncA", "yHp", "yHs",
                      "yHver", "yHz", "yken", "ylA",
                      "ymHs", "ysA", "z", "zsHn",
                      "zsHnHz", "zlAr", "yAkal", "yAkoy",
                      "yAgor"]

    """
    MorphotacticTags that can be used.
    """
    morphotactic_tags = [
        MorphologicalTag.AORIST,
        MorphologicalTag.CAUSATIVE,
        MorphologicalTag.ASIF,
        MorphologicalTag.LY,
        MorphologicalTag.EQUATIVE,
        MorphologicalTag.ASIF,
        MorphologicalTag.AGENT,
        MorphologicalTag.DIMENSION,
        MorphologicalTag.LOCATIVE,
        MorphologicalTag.ABLATIVE,
        MorphologicalTag.PASTTENSE,
        MorphologicalTag.PASTPARTICIPLE,
        MorphologicalTag.ASLONGAS,
        MorphologicalTag.COPULA,
        MorphologicalTag.SINCE,
        MorphologicalTag.CAUSATIVE,
        MorphologicalTag.P3SG,
        MorphologicalTag.PASSIVE,
        MorphologicalTag.P1SG,
        MorphologicalTag.REFLEXIVE,
        MorphologicalTag.PASSIVE,
        MorphologicalTag.P2SG,
        MorphologicalTag.P1PL,
        MorphologicalTag.ORDINAL,
        MorphologicalTag.P2PL,
        MorphologicalTag.AORIST,
        MorphologicalTag.CAUSATIVE,
        MorphologicalTag.RECIPROCAL,
        MorphologicalTag.CAUSATIVE,
        MorphologicalTag.PROGRESSIVE1,
        MorphologicalTag.A1PL,
        MorphologicalTag.A1PL,
        MorphologicalTag.RELATIVE,
        MorphologicalTag.RELATIVE,
        MorphologicalTag.ACQUIRE,
        MorphologicalTag.A3PL,
        MorphologicalTag.SINCE,
        MorphologicalTag.P3PL,
        MorphologicalTag.P3PL,
        MorphologicalTag.BECOME,
        MorphologicalTag.WITH,
        MorphologicalTag.NESS,
        MorphologicalTag.A1PL,
        MorphologicalTag.A1SG,
        MorphologicalTag.INFINITIVE2,
        MorphologicalTag.NEGATIVE,
        MorphologicalTag.ACTOF,
        MorphologicalTag.WITHOUTHAVINGDONESO,
        MorphologicalTag.INFINITIVE,
        MorphologicalTag.WITHOUTHAVINGDONESO,
        MorphologicalTag.PROGRESSIVE2,
        MorphologicalTag.NECESSITY,
        MorphologicalTag.NOTABLESTATE,
        MorphologicalTag.NARRATIVE,
        MorphologicalTag.A2SG,
        MorphologicalTag.PASSIVE,
        MorphologicalTag.DATIVE,
        MorphologicalTag.EQUATIVE,
        MorphologicalTag.LOCATIVE,
        MorphologicalTag.ABLATIVE,
        MorphologicalTag.ACCUSATIVE,
        MorphologicalTag.GENITIVE,
        MorphologicalTag.A2PL,
        MorphologicalTag.A3PL,
        MorphologicalTag.DESIRE,
        MorphologicalTag.RELATED,
        MorphologicalTag.P3SG,
        MorphologicalTag.JUSTLIKE,
        MorphologicalTag.ALMOST,
        MorphologicalTag.A2SG,
        MorphologicalTag.A2PL,
        MorphologicalTag.A3PL,
        MorphologicalTag.WITHOUT,
        MorphologicalTag.DISTRIBUTIVE,
        MorphologicalTag.CAUSATIVE,
        MorphologicalTag.DATIVE,
        MorphologicalTag.OPTATIVE,
        MorphologicalTag.ABLE,
        MorphologicalTag.FUTUREPARTICIPLE,
        MorphologicalTag.FUTURE,
        MorphologicalTag.REPEAT,
        MorphologicalTag.EVERSINCE,
        MorphologicalTag.SINCEDOINGSO,
        MorphologicalTag.NOTABLESTATE,
        MorphologicalTag.WITHOUTBEINGABLETOHAVEDONESO,
        MorphologicalTag.PRESENTPARTICIPLE,
        MorphologicalTag.BYDOINGSO,
        MorphologicalTag.FEELLIKE,
        MorphologicalTag.PASTTENSE,
        MorphologicalTag.ACCUSATIVE,
        MorphologicalTag.AGENT,
        MorphologicalTag.A1SG,
        MorphologicalTag.A2PL,
        MorphologicalTag.WHEN,
        MorphologicalTag.AFTERDOINGSO,
        MorphologicalTag.INFINITIVE3,
        MorphologicalTag.HASTILY,
        MorphologicalTag.A1PL,
        MorphologicalTag.WHILE,
        MorphologicalTag.INSTRUMENTAL,
        MorphologicalTag.NARRATIVE,
        MorphologicalTag.CONDITIONAL,
        MorphologicalTag.A3SG,
        MorphologicalTag.A2SG,
        MorphologicalTag.A2PL,
        MorphologicalTag.A3PL,
        MorphologicalTag.STAY,
        MorphologicalTag.START,
        MorphologicalTag.REPEAT]

    __meta_morpheme_list: list
    __root: Word

    def __init__(self, parse=None):
        """
        A constructor of MetamorphicParse class which creates an list metaMorphemeList which has split words
        according to +.

        PARAMETERS
        ----------
        parse : str
            String to parse.
        """
        if parse is not None:
            self.__meta_morpheme_list = []
            if parse == "+":
                self.__root = Word("+")
            else:
                words = parse.split("\\+")
                self.__root = Word(words[0])
                for i in range(1, len(words)):
                    self.__meta_morpheme_list.append(words[i])

    def getMetaMorphemeTag(self, tag: str) -> list:
        """
        The getMetaMorphemeTag method takes a String tag as an input and takes the first char of the tag. If first char
        is a punctuation it gets a substring from the tag. And gets the meta morphemes of this tag then adds to the
        result list.

        PARAMETERS
        ----------
        tag : str
            String to get meta morphemes from.

        RETURNS
        -------
        list
            List type result which holds meta morphemes.
        """
        result = []
        s = tag[0]
        if Word.isPunctuationSymbol(s):
            tag = tag[1:]
        for j in range(len(MetamorphicParse.meta_morphemes)):
            if tag == self.meta_morphemes[j]:
                result.append(MetamorphicParse.morphotactic_tags[j])
        return result

    def getWord(self) -> Word:
        """
        The getter method for Private Word root.

        RETURNS
        -------
        Word
            Word type root.
        """
        return self.__root

    def getMetaMorphemeTagForParse(self, parse: MorphologicalParse, tag: str) -> list:
        """
        getMetaMorphemeTagForParse method which also takes parse as an input. It also checks the morphotactic tags.

        PARAMETERS
        ----------
        parse : MorphologicalParse
            MorphologicalParse type input.
        tag : str
            String to get meta morphemes from.

        RETURNS
        -------
        list
            List type result which holds meta morphemes.
        """
        result = []
        s = tag[0]
        if Word.isPunctuationSymbol(s):
            tag = tag[1:]
        for j in range(len(MetamorphicParse.meta_morphemes)):
            if tag == self.meta_morphemes[j] and parse.containsTag(MetamorphicParse.morphotactic_tags[j]):
                result.append(MetamorphicParse.morphotactic_tags[j])
        return result

    def size(self) -> int:
        """
        The size method returns the size of the metaMorphemeList.

        RETURNS
        -------
        int
            The size of the metaMorphemeList.
        """
        return len(self.__meta_morpheme_list) + 1

    def addMetaMorphemeList(self, newTacticSet: str):
        """
        The addMetaMorphemeList method splits input String by + and add to the metaMorphemeList.

        PARAMETERS
        ----------
        newTacticSet : str
            String to add the metaMorphemeList.
        """
        tactics = newTacticSet.split("\\+")
        self.__meta_morpheme_list.extend(tactics)

    def removeMetaMorphemeFromIndex(self, index: int):
        """
        The removeMetaMorphemeFromIndex method removes the meta morpheme at given index from metaMorphemeList.

        PARAMETERS
        ----------
        index : int
            to remove from metaMorphemeList.
        """
        i = index - 1
        while i < len(self.__meta_morpheme_list):
            self.__meta_morpheme_list.pop(i)

    def getMetaMorpheme(self, index: int) -> str:
        """
        The getMetaMorpheme method gets the meta morpheme at given index.

        PARAMETERS
        ----------
        index : int
            is used to get the meta morpheme.

        RETURNS
        -------
        str
            metaMorphemeList's corresponding meta morpheme.
        """
        if index == 0:
            return self.__root.getName()
        else:
            return self.__meta_morpheme_list[index - 1]

    def __str__(self) -> str:
        """
        Overridden __str__ method to return resulting meta morphemes in metaMorphemeList.

        RETURNS
        -------
        str
            String result.
        """
        result = self.__root.getName()
        for meta_morpheme in self.__meta_morpheme_list:
            result = result + "+" + meta_morpheme
        return result



from __future__ import annotations
from enum import Enum, auto


class MorphologicalTag(Enum):

    """
    Noun : Alengir
    """
    NOUN = auto()
    """
    Adverb : Alelacele
    """
    ADVERB = auto()
    """
    Adjective : Alengirli
    """
    ADJECTIVE = auto()
    """
    Verb : Alıkoy
    """
    VERB = auto()
    """
    1st person singular agreement : Ben gelirim
    """
    A1SG = auto()
    """
    2nd person singular agreement : Sen gelirsin
    """
    A2SG = auto()
    """
    3rd person singular agreement : O gelir
    """
    A3SG = auto()
    """
    1st person plural agreement : Biz geliriz
    """
    A1PL = auto()
    """
    2nd person plural agreement : Siz gelirsiniz
    """
    A2PL = auto()
    """
    3rd person plural agreement : Onlar gelirler
    """
    A3PL = auto()
    """
    1st person singular possessive : Benim
    """
    P1SG = auto()
    """
    2nd person singular possessive :Senin
    """
    P2SG = auto()
    """
    3rd person singular possessive : Onun
    """
    P3SG = auto()
    """
    1st person plural possessive :  Bizim
    """
    P1PL = auto()
    """
    2nd person plural possessive : Sizin
    """
    P2PL = auto()
    """
    3rd person plural possessive : Onların
    """
    P3PL = auto()
    """
    Proper noun : John
    """
    PROPERNOUN = auto()
    """
    None possessive : Ev
    """
    PNON = auto()
    """
    Nominative Case : Kedi uyuyor.
    """
    NOMINATIVE = auto()
    """
    With : Kalemle
    """
    WITH = auto()
    """
    Without : Dikişsiz
    """
    WITHOUT = auto()
    """
    Accusatıve : Beni
    """
    ACCUSATIVE = auto()
    """
    Dative case : Bana
    """
    DATIVE = auto()
    """
    Genitive : Benim
    """
    GENITIVE = auto()
    """
    Ablative : Okuldan
    """
    ABLATIVE = auto()
    """
    Perosnal pronoun : O
    """
    PERSONALPRONOUN = auto()
    """
    Zero Derivation : Kırmızıydı
    """
    ZERO = auto()
    """
    Ability = auto() possibility : Olabilir
    """
    ABLE = auto()
    """
    Negative : Yapama
    """
    NEGATIVE = auto()
    """
    Past tense : Gitti
    """
    PASTTENSE = auto()
    """
    Conjunction or disjunction : Ama = auto() ise
    """
    CONJUNCTION = auto()
    """
    Determiner : Birtakım
    """
    DETERMINER = auto()
    """
    Duplication : Çıtır çıtır
    """
    DUPLICATION = auto()
    """
    Interjection : Agucuk
    """
    INTERJECTION = auto()
    """
    Number : bir
    """
    NUMBER = auto()
    """
    Post posıtıon : Atfen
    """
    POSTPOSITION = auto()
    """
    Punctuation : +
    """
    PUNCTUATION = auto()
    """
    Question : Mı
    """
    QUESTION = auto()
    """
    Agent : Toplayıcı
    """
    AGENT = auto()
    """
    By doing so : Zıplayarak
    """
    BYDOINGSO = auto()
    """
    Cardinal : yüz = auto() bin
    """
    CARDINAL = auto()
    """
    Causative Form : Pişirmek
    """
    CAUSATIVE = auto()
    """
    Demonstrative pronoun : Bu = auto() şu
    """
    DEMONSTRATIVEPRONOUN = auto()
    """
    Distributive : altışar
    """
    DISTRIBUTIVE = auto()
    """
    Fit for : Ahmetlik
    """
    FITFOR = auto()
    """
    Future participle : Gülecek
    """
    FUTUREPARTICIPLE = auto()
    """
    Infinitive : Biri
    """
    INFINITIVE = auto()
    """
    Ness : Ağırbaşlılık
    """
    NESS = auto()
    """
    Ordinal Number : Altıncı
    """
    ORDINAL = auto()
    """
    Passive : Açıldı
    """
    PASSIVE = auto()
    """
    Past participle : Kırılmış
    """
    PASTPARTICIPLE = auto()
    """
    Present partıcıple : Sarılan
    """
    PRESENTPARTICIPLE = auto()
    """
    Question pronoun : Kim
    """
    QUESTIONPRONOUN = auto()
    """
    Quantitative pronoun : Each
    """
    QUANTITATIVEPRONOUN = auto()
    """
    Range : 1 - 3
    """
    RANGE = auto()
    """
    Ratio : 1/2
    """
    RATIO = auto()
    """
    Real : 1.0
    """
    REAL = auto()
    """
    Reciprocal verb : Görüşmek
    """
    RECIPROCAL = auto()
    """
    Reflexive : Kendi
    """
    REFLEXIVE = auto()
    """
    Reflexive pronoun : Kendim
    """
    REFLEXIVEPRONOUN = auto()
    """
    Time : 14:28
    """
    TIME = auto()
    """
    When : Okuyunca
    """
    WHEN = auto()
    """
    While : Gelirken
    """
    WHILE = auto()
    """
    Without having done so : Çaktırmadan
    """
    WITHOUTHAVINGDONESO = auto()
    """
    PC ablative : Başka
    """
    PCABLATIVE = auto()
    """*
    PC accusative : Takiben
    """
    PCACCUSATIVE = auto()
    """
    PC dative : İlişkin
    """
    PCDATIVE = auto()
    """
    PC genitive : Yanısıra
    """
    PCGENITIVE = auto()
    """
    PC instrumental : Birlikte
    """
    PCINSTRUMENTAL = auto()
    """
    PC nominative
    """
    PCNOMINATIVE = auto()
    """
    Acquire : Kazanılan
    """
    ACQUIRE = auto()
    """
    Act of : Aldatmaca
    """
    ACTOF = auto()
    """
    After doing so : Yapıp
    """
    AFTERDOINGSO = auto()
    """
    Almost : Dikensi
    """
    ALMOST = auto()
    """
    As : gibi
    """
    AS = auto()
    """
    As if : Yaşarmışcasına
    """
    ASIF = auto()
    """
    Become : Abideleş
    """
    BECOME = auto()
    """
    Ever since : Çıkagel
    """
    EVERSINCE = auto()
    """
    Projection : Öpülesi
    """
    FEELLIKE = auto()
    """
    Hastility : Yapıver
    """
    HASTILY = auto()
    """
    In between : Arasında
    """
    INBETWEEN = auto()
    """
    Just like : Destansı
    """
    JUSTLIKE = auto()
    """
    -LY : Akıllıca
    """
    LY = auto()
    """
    Related to : Davranışsal
    """
    RELATED = auto()
    """
    Continuous : Yapadur
    """
    REPEAT = auto()
    """
    Since doing so : Amasyalı
    """
    SINCE = auto()
    """
    Since doing so : Amasyalı
    """
    SINCEDOINGSO = auto()
    """
    Start : Alıkoy
    """
    START = auto()
    """
    Stay : Bakakal
    """
    STAY = auto()
    """
    Equative : Öylece
    """
    EQUATIVE = auto()
    """
    Instrumental : Kışın = auto() arabayla
    """
    INSTRUMENTAL = auto()
    """
    Aorist Tense : Her hafta sonunda futbol oynarlar.
    """
    AORIST = auto()
    """
    Desire/Past Auxiliary : Çıkarsa
    """
    DESIRE = auto()
    """
    Future : Yağacak
    """
    FUTURE = auto()
    """
    Imperative : Otur!
    """
    IMPERATIVE = auto()
    """
    Narrative Past Tense : Oluşmuş
    """
    NARRATIVE = auto()
    """
    Necessity : Yapmalı
    """
    NECESSITY = auto()
    """
    Optative : Doğanaya
    """
    OPTATIVE = auto()
    """
    Past tense : Gitti
    """
    PAST = auto()
    """
    Present partıcıple : Sarılan
    """
    PRESENT = auto()
    """
    Progressive : Görüyorum
    """
    PROGRESSIVE1 = auto()
    """
    Progressive : Görmekteyim
    """
    PROGRESSIVE2 = auto()
    """
    Conditional : Gelirse
    """
    CONDITIONAL = auto()
    """
    Copula : Mavidir
    """
    COPULA = auto()
    """
    Positive : Gittim
    """
    POSITIVE = auto()
    """
    Pronoun : Ben
    """
    PRONOUN = auto()
    """
    Locative : Aşağıda
    """
    LOCATIVE = auto()
    """
    Relative : Gelenin
    """
    RELATIVE = auto()
    """
    Demonstrative : Bu
    """
    DEMONSTRATIVE = auto()
    """
    Infinitive2 : Gitme
    """
    INFINITIVE2 = auto()
    """
    Infinitive3 : Gidiş
    """
    INFINITIVE3 = auto()
    """
    Sentence beginning header
    """
    BEGINNINGOFSENTENCE = auto()
    """
    Sentence ending header
    """
    ENDOFSENTENCE = auto()
    """
    Title beginning header
    """
    BEGINNINGOFTITLE = auto()
    """
    Title ending header
    """
    ENDOFTITLE = auto()
    """
    Document beginning header
    """
    BEGINNINGOFDOCUMENT = auto()
    """
    Document ending header
    """
    ENDOFDOCUMENT = auto()
    """
    As long as : Yaşadıkça
    """
    ASLONGAS = auto()
    """
    Adamantly
    """
    ADAMANTLY = auto()
    """
    Percent : 15%
    """
    PERCENT = auto()
    """
    Without being able to have done so: kararlamadan
    """
    WITHOUTBEINGABLETOHAVEDONESO = auto()
    """
    Dimension : Küçücük
    """
    DIMENSION = auto()
    """
    Notable state : Anlaşılmazlık
    """
    NOTABLESTATE = auto()
    """
    Fraction : 1/2
    """
    FRACTION = auto()
    """
    Hash tag : #
    """
    HASHTAG = auto()
    """
    E-mail : @
    """
    EMAIL = auto()
    """
    Date : 11/06/2018
    """
    DATE = auto()
    """
    Code : i7-9700K
    """
    CODE = auto()
    """
    Metric : 6cmx7cmx8cm
    """
    METRIC = auto
    """
    Polite : yapınız, gidiniz
    """
    POLITE = auto
    """
    Urge : baksana, yapsana
    """
    URGE = auto



import xml.etree.ElementTree

from MorphologicalAnalysis.State import State
from MorphologicalAnalysis.Transition import Transition


class FiniteStateMachine:
    __states: list
    __transitions: dict

    def __init__(self, fileName: str):
        """
        Constructor reads the finite state machine in the given input file. It has a NodeList which holds the states
        of the nodes and there are 4 different type of nodes; stateNode, root Node, transitionNode and withNode.
        Also there are two states; state that a node currently in and state that a node will be in.

        XMLParser is used to parse the given file. Firstly it gets the document to parse, then gets its elements by the
        tag names. For instance, it gets states by the tag name 'state' and puts them into an ArrayList called
        stateList.
        Secondly, it traverses this stateList and gets each Node's attributes. There are three attributes; name, start,
        and end which will be named as states. If a node is in a startState it is tagged as 'yes', otherwise 'no'.
        Also, if a node is in a startState, additional attribute will be fetched; originalPos that represents its
        original part of speech.

        At the last step, by starting rootNode's first child, it gets all the transitionNodes and next states called
        toState then continue with the nextSiblings. Also, if there is no possible toState, it prints this case and the
        causative states.

        PARAMETERS
        ----------
        fileName : str
            the resource file to read the finite state machine. Only files in resources folder are supported.
        """
        self.__transitions = {}
        self.__states = []
        root = xml.etree.ElementTree.parse(fileName).getroot()
        for state_node in root:
            state_name = state_node.attrib["name"]
            start_state = state_node.attrib["start"] == "yes"
            end_state = state_node.attrib["end"] == "yes"
            if start_state:
                original_pos = state_node.attrib["originalpos"]
                self.__states.append(State(state_name, True, end_state, original_pos))
            else:
                self.__states.append(State(state_name, False, end_state))
        for state_node in root:
            if "name" in state_node.attrib:
                state_name = state_node.attrib["name"]
                state = self.getState(state_name)
                for transition_node in state_node:
                    state_name = transition_node.attrib["name"]
                    if "transitionname" in transition_node.attrib:
                        with_name = transition_node.attrib["transitionname"]
                    else:
                        with_name = None
                    if "topos" in transition_node.attrib:
                        root_to_pos = transition_node.attrib["topos"]
                    else:
                        root_to_pos = None
                    to_state = self.getState(state_name)
                    if to_state is not None:
                        for with_node in transition_node:
                            if "name" in with_node.attrib:
                                with_name = with_node.attrib["name"]
                                if "topos" in with_node.attrib:
                                    to_pos = with_node.attrib["topos"]
                                else:
                                    to_pos = None
                            else:
                                to_pos = None
                            if to_pos is None:
                                if root_to_pos is None:
                                    self.addTransition(fromState=state,
                                                       toState=to_state,
                                                       _with=with_node.text,
                                                       withName=with_name)
                                else:
                                    self.addTransition(fromState=state,
                                                       toState=to_state,
                                                       _with=with_node.text,
                                                       withName=with_name,
                                                       toPos=root_to_pos)
                            else:
                                self.addTransition(fromState=state,
                                                   toState=to_state,
                                                   _with=with_node.text,
                                                   withName=with_name,
                                                   toPos=to_pos)

    def isValidTransition(self, transition: str) -> bool:
        """
        The isValidTransition loops through states ArrayList and checks transitions between states. If the actual
        transition equals to the given transition input, method returns true otherwise returns false.

        PARAMETERS
        ----------
        transition : str
            is used to compare with the actual transition of a state.

        RETURNS
        -------
        bool
            True when the actual transition equals to the transition input, false otherwise.
        """
        for state in self.__transitions.keys():
            for transition1 in self.__transitions[state]:
                if transition1.__str__() is not None and transition1.__str__() == transition:
                    return True
        return False

    def getStates(self) -> list:
        """
        The getStates method returns the states in the FiniteStateMachine.
        RETURNS
        -------
        list
             StateList.
        """
        return self.__states

    def getState(self, name: str) -> State:
        """
        The getState method is used to loop through the states list and return the state whose name equal
        to the given input name.

        PARAMETERS
        ----------
        name : str
            is used to compare with the state's actual name.

        RETURNS
        -------
        State
            State if found any, None otherwise.
        """
        for state in self.__states:
            if state.getName() == name:
                return state
        return None

    def addTransition(self,
                      fromState: State,
                      toState: State,
                      _with: str,
                      withName: str,
                      toPos=None):
        """
        Another addTransition method which takes additional argument; toPos and. It creates a new Transition
        with given input parameters and adds the transition to transitions list.

        PARAMETERS
        ----------
        fromState : State
            State type input indicating the previous state.
        toState : State
            State type input indicating the next state.
        _with : str
            String input indicating with what the transition will be made.
        withName : str
            String input.
        toPos : str
            String input.
        """
        new_transition = Transition(_with=_with,
                                    toState=toState,
                                    withName=withName,
                                    toPos=toPos)
        if fromState in self.__transitions:
            transition_list = self.__transitions[fromState]
        else:
            transition_list = []
        transition_list.append(new_transition)
        self.__transitions[fromState] = transition_list

    def getTransitions(self, state: State) -> list:
        """
        The getTransitions method returns the transitions at the given state.

        PARAMETERS
        ----------
        state : State
            State input.

        RETURNS
        -------
        list
            Transitions at given state.
        """
        if state in self.__transitions:
            return self.__transitions[state]
        else:
            return []
